import GPy as GPy
import numpy as np
from mogp.neg_linear import NegLinear


class GPobs:
    def __init__(self, X, Y, kernel='rbf', signal_variance=1., signal_variance_fix=True, noise_variance=1.,
                 noise_variance_fix=False, mean_func=True, lengthscale=4.):
        """
        Individual Gaussian Process (GP) model that contributes to the overall cluster model

        Arguments:
            X: independent variable of GP model, must be a vector
            Y: dependent variable of GP model, must be a vector
            kernel (str): type of kernel for the covariance model
            signal_variance (float): numeric value to initialize signal variance
            signal_variance_fix (bool): boolean flag for fixing signal (kernel) variance
            noise_variance (float): numeric value to initialize noise variance
            noise_variance_fix (bool): boolean flag for fixing noise variance
            mean_func (bool): boolean flag to indicate if negative linear mean function should be used (otherwise, mean function of 0 used)
            lengthscale (float): numeric value to control smoothness of rbf kernel
        """

        # Radial basis function (RBF) kernel; also known as squared-exponential kernel
        if kernel == 'rbf':
            # WARNING: assumes signal has been standardized (divided by standard deviation) and time is in years
            C = GPy.kern.RBF(input_dim=1, variance=signal_variance, lengthscale=lengthscale)
            C.lengthscale.set_prior(GPy.priors.Gamma.from_EV(4., 9.), warning=False)
            if signal_variance_fix:
                C.variance.fix()  # signal variance is fixed to constant
            else:
                C.variance.set_prior(GPy.priors.Gamma.from_EV(1., .5), warning=False)

        # Linear kernel with bias
        elif kernel == 'linear':
            # WARNING: assumes signal has been standardized (divided by standard deviation) and time is in years
            C = GPy.kern.Linear(input_dim=1, variances=signal_variance) + GPy.kern.Bias(1)
            if signal_variance_fix:
                C.linear.variances.fix()
            else:
                C.variances.set_prior(GPy.priors.Gamma.from_EV(1., .5), warning=False)

        # Set negative linear mean function
        if mean_func:
            mf = NegLinear(1, 1)
            self.model = GPy.models.GPRegression(X, Y, kernel=C, mean_function=mf, noise_var=noise_variance)
            self.model.mean_function.set_prior(GPy.priors.Gamma.from_EV(2 / 3, 0.2), warning=False)
        else:
            self.model = GPy.models.GPRegression(X, Y, kernel=C, noise_var=noise_variance)

        if noise_variance_fix:
            self.model.likelihood.variance.fix()
        else:
            self.model.likelihood.variance.set_prior(GPy.priors.Gamma.from_EV(0.75, 0.25**2), warning=False)

        self.ll = self.model.log_likelihood()
        self.X = X
        self.Y = Y
        self.tempX = None
        self.tempY = None
        self.templl = None

    def calc_ll(self):
        """ log probability of a partition """
        self.ll = self.model.log_likelihood()
        return self.ll

    def update_data(self, X, Y):
        """update_data is useful when removing data from the cluster. X and Y are the new data"""
        self.X = X
        self.Y = Y

        self.model.set_XY(X=self.X, Y=self.Y)

    def calc_score(self, Xnew, Ynew):
        """calculate log predictive density, with new sample"""
        self.tempX = np.vstack([self.X, Xnew])
        self.tempY = np.vstack([self.Y, Ynew])
        temp_metadata = None
        score = np.sum(self.model.log_predictive_density(Xnew, Ynew, temp_metadata))
        return score

    def get_pred_at_loc(self, Xnew, index=0):
        """predict ynew at xnew[index]; used for sampler initializing"""
        return self.model.predict(Xnew[index].reshape(-1, 1))

    def update_suffstats(self, optimize_params=True):
        """update sufficient statistics"""
        self.X = self.tempX
        self.Y = self.tempY

        self.model.set_XY(X=self.X, Y=self.Y)
        if optimize_params:
            self.model.optimize()



import numpy as np
from GPy.mappings import Linear


class NegLinear(Linear):
    """
    A negative linear mapping

    .. math::

       F(\mathbf{x}) = -\mathbf{A} \mathbf{x})

    Arguments:
        input_dim (int): dimension of input
        output_dim (int): dimension of output
        name(str): name of mapping
    """

    def __init__(self, input_dim, output_dim, name='neg_linmap'):
        super(NegLinear, self).__init__(input_dim=input_dim, output_dim=output_dim, name=name)

    def f(self, X):
        return np.dot(X, -self.A)

    def update_gradients(self, dL_dF, X):
        self.A.gradient = np.dot(-X.T, dL_dF)

    def gradients_X(self, dL_dF):
        return np.dot(dL_dF, -self.A.T)



import numpy as np
import matplotlib

###########################################################################
#######         Functions to Interact with MoGP Model               #######
###########################################################################


def rank_cluster_prediction(model, x_new, y_new):
    """Rank clusters with new x/y data by log likelihood

    Arguments:
        model (mogp.MoGP_constrained): model to use for cluster prediction
        x_new (np.array): explanatory variable (time since symptom onset) for new patient (length: # time points)
        y_new (np.aray): responses (clinical scores) for new patient (length: # time points)

    Returns:
        rank_cluster (np.array): sorted array with most likely cluster prediction first
        cluster_ll (np.array): corresponding log likelihoods for clusters in rank_cluster
    """
    p = model.predict(x_new, y_new)
    act_clust_ids = np.where(model.allocmodel.Nk > 0)[0]
    p_act_clust = p[act_clust_ids]

    rank_idx = np.argsort(p_act_clust)[::-1]
    rank_cluster = act_clust_ids[rank_idx]
    cluster_ll = p_act_clust[rank_idx]

    return rank_cluster, cluster_ll


def generate_toy_data(seed=0):
    """Generate toy data with 3 clusters

    Arguments:
        seed (int): random seed to initialize data generation for x_value
    """
    np.random.seed(seed=seed)
    X_toy = np.sort(np.vstack([np.random.uniform(0.1, 7., (5, 4)), np.random.uniform(3., 6., (5, 4)),
                               np.random.uniform(0.1, 3., (5, 4)), np.random.uniform(0.1, 4., (5, 4))]))
    noise = np.random.randn(X_toy.shape[0], 1) * 4
    Y_slow = 48 / (1 + np.exp(-2 * (-X_toy + 6))) + noise
    Y_med = 48 / (1 + np.exp(-2 * (-X_toy + 3))) + noise
    Y_fast = 48 / (1 + np.exp(-2 * (-X_toy + 1))) + noise
    X = np.vstack([X_toy, X_toy, X_toy])
    Y = np.vstack([Y_slow, Y_med, Y_fast])
    return X, Y


def plot_mogp(ax, model, display_all_clusters=True, num_clust_disp=None):
    """Light wrapper for GPy model plotting functions to plot mixture model - will plot the largest clusters

    Arguments:
        ax (matplotlib.axes): axes for plot
        model (mogp.MoGP_constrained): model to visualize
        display_all_clusters (bool): flag indicating if all clusters should be displayed
        num_clust_disp (int): if do not want all clusters, provide the number of clusters to display

    Returns:
        ax (matplotlib.axes): axes for plot
    """
    nc = len(np.where(model.allocmodel.Nk > 0)[0])
    idx = np.argsort(-model.allocmodel.Nk)[0:nc]

    if display_all_clusters is False:
        assert num_clust_disp is not None, 'provide number of clusters to display'
        disp_clust = num_clust_disp
    else:
        disp_clust = nc

    cmap = matplotlib.cm.get_cmap('rainbow')
    colors = [cmap(i/disp_clust) for i in range(disp_clust)]

    for i, k in enumerate(idx[:disp_clust]):
        k = idx[i]
        num_pat = 'n = {}'.format(model.allocmodel.Nk[k])
        _ = model.obsmodel[k].model.plot_mean(ax=ax, label=num_pat, color=colors[i])
        _ = model.obsmodel[k].model.plot_confidence(ax=ax, label='_nolegend_', color=colors[i])
    return ax
